#include <ikra.h>

//////////////////////////////////////////////////////////////////////

inline int kernel_initialize1body(int x) {
  return x;
}

inline void kernel_initialize1(int* ptr, int size) {
  for (int i = 0; i < size; i++)
    ptr[i] = kernel_initialize1body(i);
}

VALUE kernel_initialize1wrap(VALUE self, 
			     VALUE wrapped_array,
			     VALUE _size) {
  kernel_initialize1(get_parray(wrapped_array),
		     NUM2INT(_size));

  return Qnil;
}

// This function corresponds to the Ruby expression in a block body.
inline float __ikra_reduce_kernel_2body(float x, float y) {
  return x+y;
}

// This function is an C implementation of reduction, which can be
// replaced with other implementations on other parallel platforms.
inline float __ikra_reduce_kernel_2(int* ptr, int size, float zero) {
  float accum = zero;
  for (int i = 0; i < size; i++)
    accum = __ikra_reduce_kernel_2body(accum,ptr[i]);
  return accum;
}

// This function is a wrapper that converting values between Ruby and
// C.
VALUE reduce2wrapper(VALUE self, 
		     VALUE wrapped_array,
		     VALUE _size,
		     VALUE _unit) {
  return rb_float_new
    (__ikra_reduce_kernel_2
     (get_parray(wrapped_array),
      NUM2INT(_size),
      NUM2DBL(_unit)));
}

void init_user_defined_kernels() {
  /* The following will create a module Ikra_Kernels__ in Ruby.
     Since the names of modules and classes must start with an
     uppercase letter, we can start the name with an underscore. */
  VALUE cKernel = rb_define_module("Ikra_Kernels__");

  /* The following will define a function new_kernel_1 in
     Ikra_Kernel__. */
  rb_define_module_function
    (cKernel,			/* the module to be defined in */
     "initialize1",		/* the name of the function */
     /* the pointer to the C function; the cast is needed in C++ */
     (VALUE(*)(...))kernel_initialize1wrap, 
     2);			/* the number of arguments, excluding
				   the first self */

  rb_define_module_function(cKernel, "reduce2",
			    (VALUE(*)(...))reduce2wrapper, 
			    3);	/* number of arguments */

  return;
}

#ifdef __cplusplus
/* With C++, the Init function must be declared with extern "C". */
extern "C"
#endif
void Init_ikra_kernels(){
  Init_parray();
  init_user_defined_kernels();
  return;
}
