# a library that provide compatible functions with PArray
IKRA_BACKEND="interpreter"

module ArrayConverter
  def to_a_i()  PArray.ruby_to_TN(self, 1) end
  def to_a_i1() PArray.ruby_to_TN(self, 1) end
  def to_a_i2() PArray.ruby_to_TN(self, 2) end
  def to_a_i3() PArray.ruby_to_TN(self, 3) end
  def to_a_i4() PArray.ruby_to_TN(self, 4) end
  def to_a_i5() PArray.ruby_to_TN(self, 5) end
  def to_a_i6() PArray.ruby_to_TN(self, 6) end
  def to_a_i7() PArray.ruby_to_TN(self, 7) end
  def to_a_i8() PArray.ruby_to_TN(self, 8) end
  def to_a_i9() PArray.ruby_to_TN(self, 9) end
  def to_a_f()  PArray.ruby_to_TN(self, 1) end
  def to_a_f1() PArray.ruby_to_TN(self, 1) end
  def to_a_f2() PArray.ruby_to_TN(self, 2) end
  def to_a_f3() PArray.ruby_to_TN(self, 3) end
  def to_a_f4() PArray.ruby_to_TN(self, 4) end
  def to_a_f5() PArray.ruby_to_TN(self, 5) end
  def to_a_f6() PArray.ruby_to_TN(self, 6) end
  def to_a_f7() PArray.ruby_to_TN(self, 7) end
  def to_a_f8() PArray.ruby_to_TN(self, 8) end
  def to_a_f9() PArray.ruby_to_TN(self, 9) end
end

class Array
  def self.new_(*sizes,&block)
    PArray.new(*sizes,&block)
  end

  def neighbor3(pad)
    Neighbor3.new(self,pad)
  end
  def neighbor9(pad)
    Neighbor9.new(self,pad)
  end
  def neighbor27(pad)
    Neighbor27.new(self,pad)
  end
  def neighborE3()
    NeighborE3.new(self,nil)
  end
  def neighborE9()
    NeighborE9.new(self,nil)
  end
  def neighborE27()
    NeighborE27.new(self,nil)
  end

  def sizes() [size] end

  include ArrayConverter

end

class PArray < Array
  def sizes
    @sizes
  end
  def _offset(indices)
    @sizes.zip(indices).reduce(0){ |i,(s,j)| i*s+j }
  end

  def _inc(is)
    for j in (is.length-1).downto(0)
      is[j] += 1
      if is[j] < @sizes[j]
        return true
      else
        is[j]=0
      end
    end
    false
  end

  def _loop
    is = @sizes.map{ 0 }
    @total_size.times do #while true
      yield(_offset(is),is)
      if !_inc(is)
        break
      end
    end
  end

  # to compute the numbers of elements contained in lower dimensions
  # of an array of given sizes.
  # e.g., self.rest_sizes([2,3,4]) = [2*3*4,3*4,4,1] = [24,12,4,1]
  # def self.rest_sizes(sizes)
  #   sizes.reverse.reduce([1]){ |(s,*sizes), dim | [dim*s, s, *sizes] }
  # end

  def initialize(*sizes,&block)
    @sizes = sizes
    @total_size = @sizes.reduce(&:*)
    super(@total_size,&nil)
    if block
      _loop do |ind, is|
        self[ind] = yield(*is)
      end
    end
  end

  def map
    self.class.new(*@sizes) do |*indices|
      yield(self[*indices])
    end
  end

  def [](*indices)
    fetch(_offset(indices))
  end

  def sizes() @sizes end

  def self.array_ref(ruby_array, sizes, indices, default)
    sizes.zip(indices).reduce( ruby_array ){ | array, (size,index) |
      if index < array.length then array[index] else return default end
    }
  end

  def self.array_size_of(ruby_array, dim)
    if dim==0
      ruby_array.length
    else
      ruby_array.map{ |array_sub| 
        self.array_size_of(array_sub, dim-1)
      }.max
    end
  end

  def self.ruby_to_TN(ruby_array, rank)
#    puts "ruby_to_TN(#{ruby_array}, #{rank})"
    # (max) sizes of each dimension
    sizes = Array.new(rank){ |dim| self.array_size_of(ruby_array, dim) }
    # the number of elements of the trailing dimensions
    total_size, *rest_sizes = sizes.reverse.
      reduce( [1] ){ | (s,*sizes) , dim | [dim*s, s, *sizes] }
    PArray.new(*sizes){ |*indices|
      self.array_ref(ruby_array, sizes, indices, 0)
    }
  end

  include ArrayConverter

end

class Neighbor
  def initialize(original, pad)
    @original = original
    @pad = pad
  end

  def map()
    sizes = @original.sizes
    PArray.new(*sizes) { |*idx|
      yield(make_tuple(@original, idx, sizes, @pad))
    }
  end

  def safe_access(array, index, offsets, size, pad)
    ii = index.zip(offsets).map{ |i,o| i+o }
    ii.zip(size).all?{ |i,s| 0 <= i && i < s} ?
    array[*ii] :
      ( pad==:edge ?
        array[* ii.zip(size).map{ |i,s| [[0,i].max,s-1].min }] :
        pad )
  end
end

class Neighbor3 < Neighbor
  def make_tuple(a,is,ss,pad)
    [safe_access(a, is, [-1], ss, pad),
     safe_access(a, is, [ 0], ss, pad),
     safe_access(a, is, [ 1], ss, pad)]
  end
end

class Neighbor9 < Neighbor
  def make_tuple(a,idx,ss,pad)
    N9.new{ |i,j|
      safe_access(a,idx,[i,j],ss,pad)
    }
  end
end

class Neighbor27 < Neighbor
  def make_tuple(a,idx,ss,pad)
    N27.new{ |i,j,k|
      safe_access(a,idx,[i,j,k],ss,pad)
    }
  end
end

class NeighborE3 < Neighbor
  def make_tuple(a,is,ss,pad)
    index = is[0]
    size = ss[0]
    [index==0 ? a[0] : a[index-1],
     a[index],
     index==size ? a[index-1] : a[index]]
  end
end

class NeighborE9 < Neighbor
  def make_tuple(a,idx,ss,pad)
    N9.new{ |i,j|
      i += idx[0]; j += idx[1]
      i = i==-1 ? 0 : (i == ss[0] ? i-1 : i)
      j = j==-1 ? 0 : (j == ss[1] ? j-1 : j)
      @original[i,j]
#      safe_access(a,idx,[i,j],ss,pad)
    }
  end
end

class NeighborE27 < Neighbor
  def make_tuple(a,idx,ss,pad)
    N27.new{ |i,j,k|
      i = i==-1 ? 0 : (i == ss[0] ? i-1 : i)
      j = j==-1 ? 0 : (j == ss[1] ? j-1 : j)
      k = k==-1 ? 0 : (k == ss[2] ? k-1 : k)
      a[i,j,k]
#      safe_access(a,idx,[i,j,k],ss,pad)
    }
  end
end

class N9
  def initialize()
    @a = Array.new(3){ |i|
      Array.new(3) { |j|
        yield(i-1,j-1)
      }
    }
  end

  def [](i,j) @a[i+1][j+1] end

end

class N27
  def initialize()
    @a = Array.new(3){ |i|
      Array.new(3) { |j|
        Array.new(3) { |k|
        yield(i-1,j-1,k-1)
        }
      }
    }
  end

  def [](i,j,k) @a[i+1][j+1][k+1] end

end
