type variable = string
type type_decl = variable * Ikra_type.t
val string_of_type_decls :
  (Ikra_type.t -> string) -> type_decl list -> string
type parameter_pattern =
    Var_pattern of string * Ikra_type.t
  | Array_pattern of parameter_pattern list
val empty_pattern : parameter_pattern
type literal =
    Integer_literal of int
  | Float_literal of float
  | String_literal of string
  | Symbol_literal of string
  | Range_literal of int * int
type ast_body =
    Call of tast * string * tast list * tast option
  | Literal of literal
  | Constant_assign of string * tast
  | Constant of string
  | Local_assign of string * tast
  | Local_var of string
  | Block of tast list
  | Proc of parameter_pattern * tast
  | Method_def of string * parameter_pattern * tast
  | Yield of tast
  | Array of tast list
  | While of tast * tast
  | If of tast * tast * tast option
  | For of string * tast * tast
  | Return of tast option
  | Type_decl of type_decl list * tast list
and tast = ast_body * Ikra_type.t
val is_constant : tast -> bool
val is_constant_def : string -> tast -> bool
val is_method_def : string -> tast -> bool
val method_def_body : tast -> tast
module Untyped_constructor :
  sig
    val var_pattern : string -> parameter_pattern
    val array_pattern : parameter_pattern list -> parameter_pattern
    val empty_pattern : parameter_pattern
    val call :
      tast * string * tast list * tast option -> tast
    val literal : literal -> tast
    val constant_assign : string * tast -> tast
    val constant : string -> tast
    val local_assign : string * tast -> tast
    val local_var : string -> tast
    val block : tast list -> tast
    val proc : parameter_pattern * tast -> tast
    val method_def :
      string * parameter_pattern * tast -> tast
    val yield : tast -> tast
    val array : tast list -> tast
    val while_ : tast * tast -> tast
    val if_ : tast * tast * tast option -> tast
    val for_ : string * tast * tast -> tast
    val return : tast -> tast
    val return0 : unit -> tast
    val return_o : tast option -> tast
    val type_decl : type_decl list * tast list -> tast
  end
(* val pseudo_infix_alist : (string * string) list *)
val implicit_self_name : string
val implicit_self : tast
val is_implicit_self : tast -> bool
val is_self : tast -> bool
(* val pseudo_infix_revalist : (string * string) list *)
val pseudo_name_of : string -> string
(* val is_pseudo_infix : string -> bool *)
(* val pseudo_infix : string -> string *)
val in_paren : string -> string
(* val map_concat : string -> ('a -> string) -> 'a list -> string *)
val string_of_literal : literal -> string
(* val string_of_block_parameter_sub : *)
(*   ?show_type:bool -> parameter_pattern -> string *)
(* val string_of_block_parameter : *)
(*   ?show_type:bool -> parameter_pattern -> string *)
val string_of : ?hide_type_decl:bool -> ?show_type:bool -> tast -> string
val mlexp_of : (Ikra_type.t -> string) -> tast -> string
val variables_in_pattern : parameter_pattern -> string list
module VarSet :
  sig
    type elt = String.t
    type t = Set.Make(String).t
    val empty : t
    val is_empty : t -> bool
    val mem : elt -> t -> bool
    val add : elt -> t -> t
    val singleton : elt -> t
    val remove : elt -> t -> t
    val union : t -> t -> t
    val inter : t -> t -> t
    val diff : t -> t -> t
    val compare : t -> t -> int
    val equal : t -> t -> bool
    val subset : t -> t -> bool
    val iter : (elt -> unit) -> t -> unit
    val fold : (elt -> 'a -> 'a) -> t -> 'a -> 'a
    val for_all : (elt -> bool) -> t -> bool
    val exists : (elt -> bool) -> t -> bool
    val filter : (elt -> bool) -> t -> t
    val partition : (elt -> bool) -> t -> t * t
    val cardinal : t -> int
    val elements : t -> elt list
    val min_elt : t -> elt
    val max_elt : t -> elt
    val choose : t -> elt
    val split : elt -> t -> t * bool * t
  end
val string_of_varset : VarSet.t -> string
(* val variable_set_in_pattern : parameter_pattern -> VarSet.t *)
(* val map_union : ('a -> VarSet.t) -> 'a list -> VarSet.t *)
val free_variables : tast -> VarSet.t
(* val maybe_free_variables : tast option -> VarSet.t *)
(* val recognize_paratemer_pattern : parameter_pattern -> parameter_pattern *)
val ikra_type_of_ast : tast -> Ikra_type.t
(* val remove_first : ('a -> bool) -> 'a list -> 'a * 'a list *)
(* val return_type_of_body : tast list -> (Ikra_type.t * tast list) option *)
(* val lookup_in_tdecl : *)
(*   variable -> Ikra_type.t -> type_decl list -> Ikra_type.t *)
(* val replace_types : tast list -> parameter_pattern -> parameter_pattern *)
(* val maybe : ('a -> 'b) -> 'a option -> 'b option *)
val recognize : tast -> tast
val is_ikra_type : tast -> bool


