(* configurations of the Ikra compiler, including loctaions of the
   Ruby interpreter, Ikra libraties, and so forth. *)

exception Configuration_error of string

(* unless explicitly specified, the module looks for configuration
   file with this name.  *)
let default_config_file_name = ".ikra.conf"

let config : (string*string) list ref =
  ref [
    "RUBY", "ruby";
  ]

let initialize_with : string -> unit =

  let read_lines_list : string -> string list =
  (* returns a list of strings in the given filename *)

    fun filename ->
      let inc = open_in filename in
      let rec loop lines =
	try loop (input_line inc::lines)
	with End_of_file -> 
	  close_in inc; (List.rev lines) in
      loop [] in


  let comment_line_pattern =
    Str.regexp "^[ \t]*\\(#.*\\)?$" in
  let config_line_pattern = 
    Str.regexp "^[ \t]*\\([A-Za-z_0-9]+\\)[ \t]*=[ \t]*\"\\(.*\\)\"[ \t]*$" in

  let read_as_alist path =
    let lines = read_lines_list path in
    List.concat
      (List.map (fun l ->
        if Str.string_match comment_line_pattern l 0
        then []
        else if Str.string_match config_line_pattern l 0
        then let key = Str.matched_group 1 l in
             let value = Str.matched_group 2 l in
             [(key, value)]
        else raise (Configuration_error ("unexpected configuration line: "^l))
       ) lines) in

  fun path ->
    let c = read_as_alist path in
    config := (c @ (!config))
  
let initialize : string option -> unit = function
  | None ->
      let config_path =
        try 
          List.find (fun path -> Sys.file_exists path) 
        (List.map (fun d -> Filename.concat d default_config_file_name)
           [Sys.getenv "HOME"; Sys.getcwd ()])
        with Not_found ->
          raise (Configuration_error ("No configuration files are found.")) in
      initialize_with config_path

  | Some path -> initialize_with path

let dump_config : unit -> string = fun _ ->
  String.concat ""
    (List.map (fun (k,v) -> k ^"="^v^"\n") !config)

let get_config : string -> string =
  fun key ->
    try List.assoc key !config 
    with Not_found ->
      raise (Configuration_error ("No configuration for " ^ key ^ ".\n" ^ 
      (dump_config ())))
        
let ikra_root_directory () = get_config "IKRA_HOME"
  (* "/home/masuhara/co/ikra0.2/" *)

let extconf_path () = Filename.concat (ikra_root_directory ())  "src/extconf.rb"
let parray_dir_relative () = "proto/parray" 
let parray_dir () = Filename.concat (ikra_root_directory ())  
  (parray_dir_relative ())
let path_to_ruby () = get_config "RUBY"
  (* fun _ -> *)
  (*   List.find Sys.file_exists  *)
  (*     ["/home/masuhara/.rvm/bin/ruby-1.9.3-head"; *)
  (*      (\* rvm use ruby-1.9.3-head beforehand *\) *)
  (*      "/home/masuhara/.rvm/bin/ruby-1.9.3-p448" ] *)
