# This file is called from convert_kernel.ml:generate_makefile for
# creating a Makefile.  The Makefile shall be used for compiling the
# kernel module that is generated by the Ikra compiler.

# The directory where PArray source code is located.  Specifically,
# we need a file in PARRAY_C_DIR/include/ikra.h.
# TODO: this should be parameterized.
#PARRAY_C_DIR = "/home/masuhara/current/ikra/v02/proto/parray"

# a standard Ruby library to generate a makefile
require 'mkmf'                  

# We use g++, which do not accept the following options.
CONFIG['warnflags'].gsub!('-Wdeclaration-after-statement','')
CONFIG['warnflags'].gsub!('-Wimplicit-function-declaration','')

CONFIG['debugflags'] += ' -DDEBUG' if $DEBUG_OPTION

if $CUDA
  # Let mkmkf regard *.cu files as the source files.  This will be
  # needed when we switched to nvcc from g++.
  CXX_EXT.push("cu"); CXX_EXT.delete("cc")
  SRC_EXT.push("cu"); SRC_EXT.delete("cc")
  
  # to compile code generated for CUDA, we need to
  # - set CXX to nvcc
  # - remove -fPIC option from CFLAGS (and CCDLFLAGS)
  # - clear warning options like -Wall
  # - remove -ggdb option
  CONFIG['CXX']="nvcc"
  CONFIG['warnflags']="-Xcompiler " + CONFIG['warnflags'].split(' ').join(',')
  CONFIG['warnflags'] += " -Xcompiler -Wno-switch-enum,-Wno-switch"
  CONFIG['CCDLFLAGS'].gsub!('-fPIC', '-Xcompiler -fPIC') 
  CONFIG['CFLAGS'].gsub!('-fPIC', '-Xcompiler -fPIC') 
  CONFIG['debugflags'].gsub!('-ggdb', '-Xcompiler -ggdb')
  #  CONFIG['optflags'].gsub!('-O3', '-O1')
  CONFIG.each do |k,v|
    v.gsub!(/-Wl,/,"-Xlinker ") if v
  end
  CONFIG['LDFLAGS'].gsub!('-rdynamic', '-Xcompiler -rdynamic')
  CONFIG['CPPFLAGS'] += ' -I/usr/apps.sp3/cuda/6.0/samples/common/inc/'
  # set compute capability 3.0 (for some reasons it does not work with
  # 3.5 on TSUBAME)
  CONFIG['CPPFLAGS'] += ' -arch compute_30 -code compute_30'
  CONFIG['LDFLAGS'] += ' -arch compute_30 -code compute_30'
  init_mkmf
end

# Let -I PARRAY_C_DIR/include is included in the compile option.
# (NOTE: this should be performed after init_mkmf.)
dir_config('',PARRAY_C_DIR);

# create_makefile(MODULE_NAME) # will generate Makefile.
