open OUnit2

let _ = Config.initialize None

let free_var_test_code : (string * (string list)) list
    (* a test case (source, [var1,var2,...]) represents that
       (Ast.free_vars parsed_source) will be a set of variables
       consisting of var1, var2, ... *)
    = [

  (* local variables inside a block should not be free variables *)
  "m(){ |x| _typedecl(Integer, :x, :y); y=x; y  }", [];

  (* self, either implicit or explicit, should not be free
     variables *)
  "m(0){ |x| _typedecl(Integer, :x); m(x); self.m(x) }", [];

]

let set_print : Ast.VarSet.t -> string = fun s -> "{" ^ 
  (String.concat "," (Ast.VarSet.elements s)) ^ 
  "}"

let make_free_var_test_case : string * (string list) -> 'a =
  let open Ast in
      fun (source, expected_vars) ->
	(Printf.sprintf "free_var_test(%s)" source) >::
	  fun _->
            match Ast.recognize
	      (Typeinf.recognize_type_decls
		 (Test_util.parse_ruby_string source)) with
		| Call(receiver,name,args,Some source),typ ->
		  let actual = Ast.free_variables source in
		  let expected = List.fold_right Ast.VarSet.add
		    expected_vars Ast.VarSet.empty in
	      (* Printf.printf "testing %s...\n" *)
	      (* 	(Ast.string_of source); *)
              assert_equal ~printer:set_print expected actual
	    | _ -> assert_failure "unexpected expression"


let testcode : string list = [
  
  "123*445";
  "true && false";
  "true || false";
  "!true";
  "not(true)";
]

let parse_unparse_parse : string -> 'a  =
  fun code -> ("parse_unparse_parse<"^code^">") >::
    (fun _ ->
      let ast = Test_util.parse_ruby_string code in
      let unparsed = Ast.string_of ast in
      let ast2 = Test_util.parse_ruby_string unparsed in
      assert_equal ~printer:(Ast.mlexp_of Ikra_type.string_of) ast ast2
    )

let suite = "ast test" >:::
  ("unparse_or">::(fun _ ->
    assert_equal ~printer:(fun s->s)
      "(true)||(false)"
      (Ast.string_of (Test_util.parse_ruby_string "true || false"))
   ))::
  ("unparse_not">::(fun _ ->
    assert_equal ~printer:(fun s->s)
      "!(true)"
      (Ast.string_of (Test_util.parse_ruby_string "!true"))
   ))::
  List.map make_free_var_test_case free_var_test_code
  @ List.map parse_unparse_parse testcode
  
let () =
  run_test_tt_main suite

