open OUnit2
open Test_util

module Convert = Convert.M(Convert_kernel)

let make_test_convert ?(cmp=(=)) source expected =
  ("covert_" ^ source)>::
    fun _ ->
      Convert.target_file_names None "test_convert" (fun gen_env ->
	let _ =
	  (try
	     Convert.generate_and_compile gen_env
	       (Ast.recognize
		  (Typeinf.recognize_type_decls
		     (Test_util.parse_ruby_string source)))
	   with Util.External_command_error(command, out, err) as e ->
	     Printf.printf "Compile error upon: %s\n\nOutput:\n%s\n\n\
			   Errors:\n%s\n\nRuby:\n%s\n\nCUDA:\n%s\n"
	       command out err 
	       (safe_read_lines (Convert_env.get_ruby_file_name gen_env))
	       (safe_read_lines (Convert_env.get_cuda_file_name gen_env));
	     raise e) in
	(try let output = Convert.run_converted gen_env [] in
             assert_equal ~printer:(fun s->s) ~cmp:cmp expected output
         with Util.External_command_error(command, out, err) as e ->
           Printf.printf "Error when run command: %s\n\nOutput:\n%s\n\n\
                           Errors:\n%s\n\nRuby:\n%s\n\nCUDA:\n%s\n"
             command out err 
             (Util.read_lines (Convert_env.get_ruby_file_name gen_env))
             (Util.read_lines (Convert_env.get_cuda_file_name gen_env));
           raise e);
        ())

let testdata_convert =
  let compare_float_strings x y =
    cmp_float ~epsilon:1e-7 (float_of_string x) (float_of_string y) in
  let m = make_test_convert in
  [

    (* the body of a PArray call will be converted to a call on a kernel
       method. *)
    m "_t(PArray.new(10){ |idx| _typedecl(Integer, :idx, :_RETURN); idx },
          [Array, 1, Integer])" "";

    m "_typedecl(Integer, :l); l=3; \
   _t(PArray.new(10){ |x| _typedecl(Integer, :x, :_RETURN); x+l },
        [Array, 1, Integer])" "";

    m "_typedecl([Array, 1, Integer],:a);\
   a=_t(PArray.new(10){ |x| _typedecl(Integer, :x, :_RETURN); x+1 },
        [Array,1,Integer]);\
   puts _t(a,[Array, 1, Integer]).reduce(0){ |x,y| \
        _typedecl(Integer, :x, :y, :_RETURN); x+y }"
      "55";

    m "_typedecl([Array, 1, Integer],:a, Integer, :v);\
   v=1;
   a=_t(PArray.new(10){ |x| _typedecl(Integer, :x, :_RETURN); x+v },\
        [Array,1,Integer]);\
   puts _t(a,[Array, 1, Integer]).reduce(0){ |x,y| \
        _typedecl(Integer, :x, :y, :_RETURN); x+y }"
      "55";

    m "_typedecl([Array, 1, Integer],:a,:b);\
   a=_t(PArray.new(10){ |x| _typedecl(Integer, :x, :_RETURN); x },\
        [Array, 1, Integer]);\
   b=_t(a,[Array, 1, Integer]).map{ |x| \
        _typedecl(Integer, :x, :_RETURN); x+1 };\
   puts _t(b,[Array, 1, Integer]).reduce(0){ |x,y| \
        _typedecl(Integer, :x, :y, :_RETURN); x+y }"
      "55";

    m "puts _t(_t(PArray.new(10){ |x| _typedecl(Integer, :x, :_RETURN); x },\
         [Array, 1, Integer]).map{ |x| _typedecl(Integer, :x, :_RETURN); x+1 },\
         [Array, 1, Integer]).\
      reduce(0){ |x,y| _typedecl(Integer, :x, :y, :_RETURN); x+y }"
      "55";

    (* test of using floating point numbers with lexical variable *)
    m ~cmp:compare_float_strings
      "_typedecl(Float, :l); l=0.1;\
   puts _t(PArray.new(10){ |x| _typedecl(Integer, :x, Float, :_RETURN); \
                               x*l },\
           [Array, 1, Float]).\
        reduce(0.0){ |x,y| _typedecl(Float, :x, :y, :_RETURN); x+y }"
      "4.5";

    (* test of using floating point numbers *)
    m ~cmp:compare_float_strings
      "puts _t(PArray.new(10){ |x| _typedecl(Integer, :x, Float, :_RETURN); \
                               x*0.1 },\
           [Array, 1, Float]).\
        reduce(0.0){ |x,y| _typedecl(Float, :x, :y, :_RETURN); x+y }"
    "4.5";

    (* test of a map that changes type *)
    m ~cmp:compare_float_strings
      "puts _t(_t(PArray.new(10){ |x| _typedecl(Integer, :x, :_RETURN); \
                               x+1 },[Array, 1, Integer]).\
                      map{ |x| _typedecl(Integer, :x, Float, :_RETURN);\
                                x*0.1 }, [Array,1,Float]).\
        reduce(0.0){ |x,y| _typedecl(Float, :x, :y, :_RETURN); x+y }"
    "5.5";

  (* test of boolean operators *)
    m "puts true && false || !true"
      "false";
  ]

(* let test_convert_runs = List.map make_test_convert testdata_convert *)
  
let suite = "suite">:::testdata_convert

let _ = Config.initialize None

let () = run_test_tt_main suite
