open OUnit2

let set_print = Type_constraints.set_print

let _ = Config.initialize None

module ExpectedTypeConstraintConstructors = struct
  open Ikra_type
  let (<) t0 t1 = Type_constraints.Constr.Less_than(t0,t1)
  let (=) t0 t1 = Type_constraints.Constr.Equal(t0,t1)
  let (=>) t0 t1 = Tproc(t0,t1,None)
  let t n = Tvar n
  let cset =
    List.fold_left (fun set con -> Type_constraints.Constraints.add con set)
      Type_constraints.Constraints.empty
  let mtype receiver name args return =
    Type_constraints.Constr.Has_method(receiver,name,Tproc(args,return,None))
end

let printer = fun (c,s) ->
  Printf.sprintf "c=%s, s=[%s]" (set_print c)
    (String.concat ","
       (List.map (fun (t1,t2) ->
         (Ikra_type.string_of t1) ^ "->" ^ (Ikra_type.string_of t2))
          s))

let cmp = fun (c1,s1) (c2,s2) ->
  Type_constraints.Constraints.equal c1 c2 && s1=s2

let test_resolve resolver (cstraints,
                  cycle_removed_expected,
                  equality_removed_expected, subst_expected,
                  cs_iteration_expected, subst_iteration_expected) =
  (set_print cstraints)>:: (fun _ ->
    let cycle_removed = (Cresolve.remove_cycle (cstraints,[])) in
    assert_equal ~printer ~cmp
      (cycle_removed_expected, []) cycle_removed;
    let equality_removed = (Cresolve.unify cycle_removed) in
    assert_equal ~printer ~cmp
      (equality_removed_expected, subst_expected) equality_removed;
    let iteration_result = resolver cstraints in
    assert_equal ~printer ~cmp
      (cs_iteration_expected, subst_iteration_expected) iteration_result
   )

let test_resolve1 = test_resolve Cresolve.resolve_one'
let test_resolve2 (cstraints,
                   cs_iteration_expected, subst_iteration_expected) =
  (set_print cstraints)>:: (fun _ ->
    let result = Cresolve.resolve cstraints in
    assert_equal ~printer ~cmp 
      (cs_iteration_expected, subst_iteration_expected) result)


let cycle_constraints =
  let open ExpectedTypeConstraintConstructors in
      let open Ikra_type in
          List.map test_resolve1
            [
              cset [ t 0 < t 1; t 1 < t 0 ],
              cset [ t 0 < t 1; t 1 < t 0 ; t 1 = t 0 ],
              cset [ t 0 < t 0; t 0 < t 0 ], [ t 1, t 0 ],
              cset [  ], [ t 1, t 0 ];

              cset [ t 0 < t 1; t 1 < t 0; t 2 < t 3; t 3 < t 2 ],
              cset [ t 0 < t 1; t 1 < t 0; t 2 < t 3; t 3 < t 2;
                     t 1 = t 0; t 3 = t 2 ],
              cset [ t 0 < t 0; t 0 < t 0; t 2 < t 2; t 2 < t 2; ],
              [ t 3, t 2; t 1, t 0;],
              cset [ ], [ t 3, t 2; t 1, t 0;];

              cset [ t 0 < t 1; t 1 < t 2; t 2 < t 0 ],
              cset [ t 0 < t 1; t 1 < t 2; t 2 < t 0 ; t 2 = t 0; t 2 = t 1 ],
              cset [ t 1 < t 1; t 1 < t 1; t 1 < t 1 ;], [ t 0, t 1; t 2, t 1 ],
              cset [ ], [ t 0, t 1; t 2, t 1 ];

              cset [ t 0 < t 1; t 0 = t 1 ],
              cset [ t 0 < t 1; t 0 = t 1 ],
              cset [ t 1 < t 1 ], [ t 0, t 1 ],
              cset [  ], [ t 0, t 1 ];

              cset [ t 2 = t 0; t 2 = t 1 ],
              cset [ t 2 = t 0; t 2 = t 1 ],
              cset [ ], [ t 0, t 1; t 2, t 1 ],
              cset [ ], [ t 0, t 1; t 2, t 1 ];

              cset [ t 0 < t 1; t 1 = t 2; t 2 < t 0 ],
              cset [ t 0 < t 1; t 1 = t 2; t 2 < t 0 ],
              cset [ t 0 < t 2; t 2 < t 0 ], [ t 1, t 2; ],
              cset [  ], [ t 2, t 0; t 1, t 0; ];

              cset [ Tarray(1,t 0) = Tarray(1,t 1); t 0 < t 2; t 2 < t 1 ],
              cset [ Tarray(1,t 0) = Tarray(1,t 1); t 0 < t 2; t 2 < t 1 ],
              cset [ t 0 = t 1; t 0 < t 2; t 2 < t 1], [ ],
              cset [  ], [ t 2, t 1; t 0, t 1; ];

              cset [ mtype (t 2) "m" [t 2; t 0] (t 2); t 2 = t 0 ],
              cset [ mtype (t 2) "m" [t 2; t 0] (t 2); t 2 = t 0 ],
              cset [ mtype (t 0) "m" [t 0; t 0] (t 0) ], [ t 2, t 0],
              cset [ mtype (t 0) "m" [t 0; t 0] (t 0) ], [ t 2, t 0];

              (* constraints generated by x=1+1 *)
              cset [ t 1 = t 0; t 2 = t 1; t 2 = t 6; t 3 < t 6;
                     mtype (t 5) "+" [t 7] (t 3);
                     t 4 < t 7; t 4 = Tint; t 5 = Tint ],
              cset [ t 1 = t 0; t 2 = t 1; t 2 = t 6; t 3 < t 6;
                     mtype (t 5) "+" [t 7] (t 3);
                     t 4 < t 7; t 4 = Tint; t 5 = Tint ],
              cset [ t 3 < t 6; mtype Tint "+" [t 7] (t 3); Tint < t 7;  ],
              [ t 5, Tint; t 4, Tint; t 0, t 6; t 2, t 6; t 1,t 6;     ],
              cset [ t 3 < t 6; mtype Tint "+" [t 7] (t 3); Tint < t 7;  ],
              [ t 5, Tint; t 4, Tint; t 0, t 6; t 2, t 6; t 1,t 6;     ];

              (* generated by x=if x then x else x end *)
              cset [ t 1 = t 0; t 2 = t 1; t 2 = t 7; t 3 < t 7; t 4 < t 3;
                     t 5 < t 3; t 6 < Tbool; t 4 = t 7; t 5 = t 7; t 6 = t 7 ],
              cset [ t 1 = t 0; t 2 = t 1; t 2 = t 7; t 3 < t 7; t 4 < t 3;
                     t 5 < t 3; t 6 < Tbool; t 4 = t 7; t 5 = t 7; t 6 = t 7 ],
              cset [ t 3 < t 7; t 7 < t 3; t 7 < t 3; t 7 < Tbool; ],
              [ t 6, t 7; t 5, t 7; t 4, t 7; t 0, t 7; t 2, t 7; t 1, t 7; ],
              cset [ t 3 < Tbool; ],
              [ t 7, t 3; t 6, t 3; t 5, t 3; t 4, t 3; t 0, t 3; t 2, t 3;
                t 1, t 3; ];

	    (* t0 = t0 should be eliminated *)
	      cset [ t 0 = t 0 ],
	      cset [ t 0 = t 0 ],	(* cycle removed *)
	      cset [  ], [],		(* equality removed *)
	      cset [  ], [];		(* after second level iteration *)

            (* transitive inequalities *)
              cset [ t 0 < t 1; t 1 < t 2 ],
              cset [ t 0 < t 1; t 1 < t 2 ],
              cset [ t 0 < t 1; t 1 < t 2 ], [],
              cset [ t 0 < t 1; t 1 < t 2; t 0 < t 2 ], [];

            ]

let second_level =
  (* domain specific constraint resolution  *)
  let open ExpectedTypeConstraintConstructors in
      let open Ikra_type in
          List.map test_resolve2
            [
              cset [ t 0 < Tint;  ],    (* input *)
              cset [  ], [ t 0, Tint ];  (* iteration w/second level *)

              cset [ t 0 < Tbool;  ],    (* input *)
              cset [  ], [ t 0, Tbool ];  (* iteration w/second level *)

              cset [ Tfloat < t 0;  ],    (* input *)
              cset [  ], [ t 0, Tfloat ];  (* iteration w/second level *)

	      (* it's rather obvious *)
              cset [ Tint < Tfloat;  ],    (* input *)
              cset [  ],[ ];  (* iteration w/second level *)

	      (* the following test means, if we know t0 is of type
	      	 Int or Float, we prefer Int. *)
              cset [ Tint < t 0;  ],    (* input *)
              cset [  ], [ t 0, Tint ];  (* iteration w/second level *)

	      (* when there is a contradiction *)
              cset [ Tfloat < Tint;  ],    (* input *)
              cset [ Tfloat < Tint;  ],[ ];  (* iteration w/second level *)

	      (* the following test means, if we know t0 is of type
	      	 Int or Float, we prefer Int. *)
              cset [ t 0 < Tfloat;  ],    (* input *)
              cset [  ], [ t 0, Tint ];  (* iteration w/second level *)

	      (* built-in rules for arithmetics. *)
              cset [ mtype Tint "+" [t 1] (t 2);  ],    (* input *)
	      (* will become t 1 < Tfloat; t 1 = t 2 and then *)
              cset [  ], [ t 2, Tint; t 1, Tint];

	      (* built-in rules for arithmetics. *)
              cset [ mtype Tint "+" [Tfloat] (t 2);  ],    (* input *)
	      (* will become Tfloat < Tfloat; Tfloat = t 2 and then *)
              cset [  ], [ t 2, Tfloat];

	      (* built-in rules for arithmetics. *)
              cset [ mtype Tint "*" [t 1] (t 2);  ],    (* input *)
	      (* will become t 1 < Tfloat; t 1 = t 2 and then *)
              cset [  ], [ t 2, Tint; t 1, Tint];

	      (* built-in rules for arithmetics. *)
              cset [ mtype Tfloat "*" [t 1] (t 2);  ],    (* input *)
	      (* will become t 1 < Tfloat; t 1 = t 2 and then *)
              cset [  ], [ t 1, Tint; t 2, Tfloat ];


	      (* built-in rules for arithmetics. *)
              cset [ mtype (t 0) "*" [Tfloat] (t 2);  Tint < t 0 ;
	      	   ],    (* input *)
	      (* will become t 1 < Tfloat; t 1 = t 2 and then *)
              cset [  ], [ t 0, Tint; t 2, Tfloat;  ];

	      (* built-in rules for arithmetics. *)
              cset [ mtype (t 0) "*" [t 1] (t 2);  Tint < t 0 ;
		     t 1 = Tfloat;],    (* input *)
	      (* will become t 1 < Tfloat; t 1 = t 2 and then *)
              cset [  ], [ t 0, Tint; t 2, Tfloat; t 1, Tfloat ];  

              (* arrays are invariant types *)
              cset [ (Tarray(1,t 0)) < (Tarray(1,t 1)) ],
              cset [ ], [t 0, t 1];

            (* neighbor support *)
              cset [ mtype (Tarray(2,Tint)) "neighbor9" [Tint]
                       (Tarray(2,Tneighbor(9,2,Tint))) ],
                     cset [], [];

            (* range operator (..) *)
              cset [
                mtype (Tobject) "__PSEUDO_dot2" [t 9; t 8] (t 3);
                ],
              cset [], [t 9, Tint; t 8, Tint; t 3, Trange];

            (* proc types should be unified *)
              cset [ ([t 0; t 1] => (t 2)) = ([Tint; Tint] => Tfloat)] ,
              cset [ ] ,[
                t 2, Tfloat;
                t 1, Tint;
                t 0, Tint;
              ];

            (* indirect subtype relation *)
              cset [ Tint < (t 23);
                     (t 6) < Tfloat;
                     (t 23) < (t 22);
                     mtype (t 22) "*" [Tfloat] (t 10);
                     mtype (t 10) "-" [Tint] (t 6);
                   ],
              cset [ ],
              [ 
                t 23, Tint;
                t 22, Tint;
                t 6,  Tfloat;
                t 10, Tfloat;
              ];

            (* float *)
              cset [ Tfloat < t 0; Tobject < t 0 ],
              cset [  ], [t 0, Tobject];

            (* FIXME: I have deleted something here.  Recover from the
               repository!!!

               cset [ Tarray(1,Tint) < Tobject ]

            *)

            ]
            

let make_test_typeinf (source, expected) =
  ("test_typeinf:"^source)>::(fun _ ->
    try
      let typed_source =
        Ast.string_of ~show_type:true
	  (Cresolve.typeinf
	     (Typeinf.add_type_vars
	        (Let_intro.let_intro
		   (Test_util.parse_ruby_string source)))) in
      assert_equal ~printer:(fun x->x) expected typed_source
    with Cresolve.Unresolved(ast2,raints2,ast3,raints3) ->
      assert_failure
        (Printf.sprintf "unresolved: Level 2:\nast=%s, \nconstraints=%s\nLevel 3:\nast=%s, \nconstraints=%s\n"
           (Ast.string_of ~show_type:true ast2)
           (set_print raints2)
           (Ast.string_of ~show_type:true ast3)
           (set_print raints3)))

let typeinf_tests = List.map make_test_typeinf [
  "x=0", "(_typedecl(int,:x);(x = ((0):int)):int):int";

  "x=1; x=x+1;", "(_typedecl(int,:x);((x = ((1):int)):int;(x = (((x):int.+((1):int)):int)):int):int):int";

  "x=1; x=x+1.0;",
  "(_typedecl(float,:x);((x = ((1):int)):float;(x = (((x):float.+((1.000000):float)):float)):float):float):float";

  "x=1; y=x+1.0;",
  "(_typedecl(int,:x,float,:y);((x = ((1):int)):int;(y = (((x):int.+((1.000000):float)):float)):float):float):float";

  "a = Array.new(100){ |i| i*i*1.0 }; b = a.map{ |v| v/2 };",
  "(_typedecl(array(1,float),:a,array(1,float),:b);((a = (((Array):class<Array>.new((100):int){( | i:int | (_typedecl(int,:i);(((i):int.*((i):int)):int.*((1.000000):float)):float):float):int->float}):array(1,float))):array(1,float);(b = (((a):array(1,float).map(){( | v:float | (_typedecl(float,:v);((v):float./((2):int)):float):float):float->float}):array(1,float))):array(1,float)):array(1,float)):array(1,float)";

  "a = Array.new(100){ |i| i*i*1.0 }; b = a.reduce(0){ |x,y| x+y };",
  "(_typedecl(array(1,float),:a,float,:b);((a = (((Array):class<Array>.new((100):int){( | i:int | (_typedecl(int,:i);(((i):int.*((i):int)):int.*((1.000000):float)):float):float):int->float}):array(1,float))):array(1,float);(b = (((a):array(1,float).reduce((0):int){( | x:float,y:float | (_typedecl(float,:x,float,:y);((x):float.+((y):float)):float):float):float*float->float}):float)):float):float):float";

  "a = Array.new(100){ |i| i*i*1 }; b = a.reduce(0){ |x,y| x+1.0 };",
  "(_typedecl(array(1,int),:a,float,:b);((a = (((Array):class<Array>.new((100):int){( | i:int | (_typedecl(int,:i);(((i):int.*((i):int)):int.*((1):int)):int):int):int->int}):array(1,int))):array(1,int);(b = (((a):array(1,int).reduce((0):int){( | x:float,y:float | (_typedecl(float,:x,float,:y);((x):float.+((1.000000):float)):float):float):float*float->float}):float)):float):float):float";

  (* NOTE we use a different name because Array.new(size,val)
     creates one dimensional array with an initial value. *)
  "a = Array.new_(100,100){ |i,j| i*j*1 }; b = a.map{ |x| x+1.0 };",
  "(_typedecl(array(2,int),:a,array(2,float),:b);((a = (((Array):class<Array>.new_((100):int,(100):int){( | i:int,j:int | (_typedecl(int,:i,int,:j);(((i):int.*((j):int)):int.*((1):int)):int):int):int*int->int}):array(2,int))):array(2,int);(b = (((a):array(2,int).map(){( | x:int | (_typedecl(int,:x);((x):int.+((1.000000):float)):float):float):int->float}):array(2,float))):array(2,float)):array(2,float)):array(2,float)";

  "puts Array.new(10){|x| x }.reduce(0){ |x,y| x+y }",
  "(puts((((Array):class<Array>.new((10):int){( | x:int | (_typedecl(int,:x);(x):int):int):int->int}):array(1,int).reduce((0):int){( | x:int,y:int | (_typedecl(int,:x,int,:y);((x):int.+((y):int)):int):int):int*int->int}):int)):object";

  "def m(x) x*1; end; m(10)",
  "((def m(x:int)(((x):int.*((1):int)):int):int; end):t1;(m((10):int)):int):int";

  "b = 0 == 1",
  "(_typedecl(bool,:b);(b = (((0):int.==((1):int)):bool)):bool):bool";

  (* zip support *)
  "a=Array.new(10){|x| x}; b=a.zip(a); b.map{|x,y| x+y}",
  "(_typedecl(array(1,int),:a,array(1,tuple<int,int>),:b);((a = (((Array):class<Array>.new((10):int){( | x:int | (_typedecl(int,:x);(x):int):int):int->int}):array(1,int))):array(1,int);(b = (((a):array(1,int).zip((a):array(1,int))):array(1,tuple<int,int>))):array(1,tuple<int,int>);((b):array(1,tuple<int,int>).map(){( | x:int,y:int | (_typedecl(int,:x,int,:y);((x):int.+((y):int)):int):int):int*int->int}):array(1,int)):array(1,int)):array(1,int)";

(* neighbor support *)
  "Array.new_(10,10){ |x,y| x*10+y }.neighbor9(0)",
  "(((Array):class<Array>.new_((10):int,(10):int){( | x:int,y:int | (_typedecl(int,:x,int,:y);(((x):int.*((10):int)):int.+((y):int)):int):int):int*int->int}):array(2,int).neighbor9((0):int)):array(2,neighbor(9,2,int))";

  "puts Array.new_(10,10){ |x,y| x*10+y }\
             .neighbor9(0).map{ |n| (n[-1,-1] + n[-1,0] + n[1,1] +\
                                     n[ 0,-1] + n[ 0,0] + n[0,1] +\
                                     n[ 1,-1] + n[ 1,0] + n[1,1]) / 9
             }.reduce(0){ |x,y| x+y }",
  "(puts((((((Array):class<Array>.new_((10):int,(10):int){( | x:int,y:int | (_typedecl(int,:x,int,:y);(((x):int.*((10):int)):int.+((y):int)):int):int):int*int->int}):array(2,int).neighbor9((0):int)):array(2,neighbor(9,2,int)).map(){( | n:neighbor(9,2,int) | (_typedecl(neighbor(9,2,int),:n);(((((((((((n):neighbor(9,2,int).[]((-1):int,(-1):int)):int.+(((n):neighbor(9,2,int).[]((-1):int,(0):int)):int)):int.+(((n):neighbor(9,2,int).[]((1):int,(1):int)):int)):int.+(((n):neighbor(9,2,int).[]((0):int,(-1):int)):int)):int.+(((n):neighbor(9,2,int).[]((0):int,(0):int)):int)):int.+(((n):neighbor(9,2,int).[]((0):int,(1):int)):int)):int.+(((n):neighbor(9,2,int).[]((1):int,(-1):int)):int)):int.+(((n):neighbor(9,2,int).[]((1):int,(0):int)):int)):int.+(((n):neighbor(9,2,int).[]((1):int,(1):int)):int)):int./((9):int)):int):int):neighbor(9,2,int)->int}):array(2,int).reduce((0):int){( | x:int,y:int | (_typedecl(int,:x,int,:y);((x):int.+((y):int)):int):int):int*int->int}):int)):object";

(* unary minus *)
  "x=0; x=-x; y=0.0; y=-y",
  "(_typedecl(int,:x,float,:y);((x = ((0):int)):int;(x = (((x):int.-@()):int)):int;(y = ((0.000000):float)):float;(y = (((y):float.-@()):float)):float):float):float";
(* global variable *)
  "x=$0",
  "(_typedecl(object,:x);(x = (($0):object)):object):object";

(* to_i *)
  "\"123\".to_i",
  "((\"123\"):object.to_i()):int";

  (* &&, ||, not *)
  (* "Array.new(512){ |n| if true && n < 2 then 1 else 0 end}", *)
  "n=1;if true && n<2 then n else 0 end",
  "(_typedecl(int,:n);((n = ((1):int)):int;(if (((true):bool)&&(((n):int.<((2):int)):bool)):bool then (n):int else (0):int end):int):int):int";

  "n=1;if false || !(true && n<2) then n else 0 end",
  "(_typedecl(int,:n);((n = ((1):int)):int;(if (((false):bool)||((!((((true):bool)&&(((n):int.<((2):int)):bool)):bool)):bool)):bool then (n):int else (0):int end):int):int):int";

  (* map on array of tuples *)
  "Array.new(1){ |x| x }.map{ |x| [x,x] }.map { |x,y| x+y }",
  "((((Array):class<Array>.new((1):int){( | x:int | (_typedecl(int,:x);(x):int):int):int->int}):array(1,int).map(){( | x:int | (_typedecl(int,:x);([(x):int,(x):int]):tuple<int,int>):tuple<int,int>):int->tuple<int,int>}):array(1,tuple<int,int>).map(){( | x:int,y:int | (_typedecl(int,:x,int,:y);((x):int.+((y):int)):int):int):int*int->int}):array(1,int)";

  "Array.new(1){ |x| [x,x] }.map { |x,y| x+y }",
  "(((Array):class<Array>.new((1):int){( | x:int | (_typedecl(int,:x);([(x):int,(x):int]):tuple<int,int>):tuple<int,int>):int->tuple<int,int>}):array(1,tuple<int,int>).map(){( | x:int,y:int | (_typedecl(int,:x,int,:y);((x):int.+((y):int)):int):int):int*int->int}):array(1,int)";

  (* Object#to_a_i, etc. *)
  "x.to_a_i()",  "((x()):object.to_a_i()):array(1,int)";
  "x.to_a_i1()", "((x()):object.to_a_i1()):array(1,int)";
  "x.to_a_i2()", "((x()):object.to_a_i2()):array(2,int)";
  "x.to_a_f()",  "((x()):object.to_a_f()):array(1,float)";
  "x.to_a_f1()", "((x()):object.to_a_f1()):array(1,float)";
  "x.to_a_f2()", "((x()):object.to_a_f2()):array(2,float)";

]

let contradiction_detection (craints, detects) =
  let craints_str = set_print craints in
  (Printf.sprintf "Contradiction_detection(%s)" craints_str) >::
    (fun _ ->
      try (let _ = Cresolve.resolve craints in
           assert_failure (Printf.sprintf "undetected contradiction in %s"
                             craints_str))
      with Type_constraints.Contradiction(c) ->
        assert_equal ~printer:Type_constraints.Constr.string_of
          detects c)  (* detection of contradiction *)

let contradiction_detections =
  let open ExpectedTypeConstraintConstructors in
  let open Ikra_type in
  List.map contradiction_detection [
    cset [ t 0 = Tint; t 0 = Tobject ], (Tint = Tobject);
    cset [ t 0 = Tbool; t 0 = Tobject ], (Tbool = Tobject);
]
  

let suite = "suite">:::
  cycle_constraints
  @ [
    "apply_to_type">::(fun _ ->
      let open ExpectedTypeConstraintConstructors in
      assert_equal  (t 1) (Cresolve.apply_to_type [ t 0, t 1 ] (t 0)));
    "apply_to_craints">::(fun _ ->
      let open ExpectedTypeConstraintConstructors in
          assert_equal ~printer:set_print
            (cset [ t 1 < t 1 ])
            (Cresolve.apply_to_craints [ t 0, t 1 ] (cset [ t 0 < t 1 ]) ));
    "apply_to_craint">::(fun _ ->
      let open ExpectedTypeConstraintConstructors in
          assert_equal
            (t 1 < t 1)
            (Cresolve.apply_to_craint [ t 0, t 1 ] (t 0 < t 1) ));
    "process_rule">::(fun _ ->
      let open ExpectedTypeConstraintConstructors in
          assert_equal
            (cset [ ], [t 0, t 1; t 2, t 1])
            (Cresolve.process_rule (t 0 = t 1)
               (cset [ ], [t 2, t 0])));

    "apply_to_craint2">::(fun _ ->
      let open ExpectedTypeConstraintConstructors in
          assert_equal
            (t 1 = t 0)
            (Cresolve.apply_to_craint [ t 2, t 1 ] (t 2 = t 0) ));

    ]
  @ second_level
  @ typeinf_tests
  @ contradiction_detections

let () =
  run_test_tt_main suite
