open OUnit2

let _ = Config.initialize None

let make_test_cuda_exp (source, expected) =
  ("statements_of_ruby_ast_" ^ source)>::
    fun _ ->
      let rexp = (Ast.recognize
		    (Typeinf.recognize_type_decls
		       (Test_util.parse_ruby_string source))) in
      assert_equal ~printer:Cuda.string_of_statement
	expected (Cuda.statement_of_ruby_ast rexp)

let make_test_exp ruby_code cuda_exp =
  let open Cuda in
      (
	"_typedecl(Integer, :x, Float, :y); x=0; y=0.0;"^ruby_code,
	Block[Vardecl("x", CUDAint);
              Vardecl("y", CUDAfloat);
	      Exp(Binary_op(Assign, Var "x", Literal(Ast.Integer_literal 0)));
	      Exp(Binary_op(Assign, Var "y", Literal(Ast.Float_literal 0.0)));
	      Exp(cuda_exp)]
      );;

let testdata_cuda_exp =
  let open Cuda in [
    make_test_exp "x+x"       (Binary_op(Add, Var "x", Var "x"));
    make_test_exp "x-x"       (Binary_op(Sub, Var "x", Var "x"));
    make_test_exp "x/x"       (Binary_op(Div, Var "x", Var "x"));
    make_test_exp "x*x"       (Binary_op(Mul, Var "x", Var "x"));

    (* modulo operator should be compiled differently according to types *)
    make_test_exp "_t(x,Integer)%x"
                              (Binary_op(Mod, Var "x", Var "x"));
    make_test_exp "_t(y,Float)%x"
                              (Funcall("fmodf", [Var "y"; Var "x"]));

    make_test_exp "x||x"      (Binary_op(Logical_or, Var "x", Var "x"));
    make_test_exp "x&&x"      (Binary_op(Logical_and, Var "x", Var "x"));
    make_test_exp "(x or x)"  (Binary_op(Logical_or, Var "x", Var "x"));
    make_test_exp "(x and x)" (Binary_op(Logical_and, Var "x", Var "x"));
    make_test_exp "x|x"       (Binary_op(Bitwise_or, Var "x", Var "x"));
    make_test_exp "x^x"       (Binary_op(Exclusive_or, Var "x", Var "x"));
    make_test_exp "x&x"       (Binary_op(Bitwise_and, Var "x", Var "x"));
    make_test_exp "x==x"      (Binary_op(Equal, Var "x", Var "x"));
    make_test_exp "x!=x"      (Unary_op(Logical_not,
					Binary_op(Equal, Var "x", Var "x")));
    make_test_exp "x<x"       (Binary_op(Less_than, Var "x", Var "x"));
    make_test_exp "x>x"       (Binary_op(Greater_than, Var "x", Var "x"));
    make_test_exp "x<=x"      (Binary_op(Less_or_equal, Var "x", Var "x"));
    make_test_exp "x>=x"      (Binary_op(Greater_or_equal,
					 Var "x", Var "x"));
    make_test_exp "x<<x"      (Binary_op(Shift_left, Var "x", Var "x"));
    make_test_exp "x>>x"      (Binary_op(Shift_right, Var "x", Var "x"));

    make_test_exp "-x"        (Unary_op(Unary_minus, Var "x"));
    make_test_exp "!x"        (Unary_op(Logical_not, Var "x"));
    make_test_exp "~x"        (Unary_op(Bitwise_not, Var "x"));
    make_test_exp "+x"        (Unary_op(Unary_plus, Var "x"));

    make_test_exp "x=x"       (Binary_op(Assign, Var "x",Var "x"));
    make_test_exp "x+=x"      (Binary_op(Assign, Var "x",
				      Binary_op(Add, Var "x", Var "x")));
    make_test_exp "x-=x"      (Binary_op(Assign, Var "x",
				      Binary_op(Sub, Var "x", Var "x")));


    make_test_exp "x[x]"      (Aref(Var "x", Var "x"));

    make_test_exp "x ? x : x" (Ternary_op(Cond,Var "x", Var "x", Var "x"));

    make_test_exp "x.to_i"    (Cast(CUDAint, Var "x"));
    make_test_exp "x**x"      (Funcall("fexp", [Var "x";Var "x"]));
  ];;

let test_cuda_exps = List.map make_test_cuda_exp testdata_cuda_exp
  

let suite =
  "suite_cuda">:::[
  ]
    @ test_cuda_exps			

let () =
  run_test_tt_main suite
