open OUnit2

(* the following two variables control sets test cases that are not
   yet supported. *)
let do_test_tuple = false
let do_test_procedure_as_symbol = false

let test_scripts = [
  "puts Array.new(10){|x| x }.reduce(0){ |x,y| x+y }";
  "puts Array.new(10){|x| x*1.0 }.reduce(0){ |x,y| x+y }";
  "puts Array.new(10){|x| x }.reduce(0.0){ |x,y| x+y }";
  "puts Array.new(10){|x| x }.map{ |v| v*1.0 }.reduce(0){ |x,y| x+y }";
  "puts Array.new(10000){|x| x }.reduce(0){ |x,y| x+y }";
  (* The converted program of the next one will produce an incorrect
     result because the result exceeds 2**31, which is rounded to 32 bit
     integers, while Ruby uses bignums. *)
  (* "puts Array.new(100000){|x| x }.reduce(0){ |x,y| x+y }"; *)

  "puts Array.new(10){|x| x }.reduce(0){ |x,y| x+y+0.0 }";
  "puts Array.new(10){|x| x }.reduce(0){ |x,y| a=x+y;a }";

  (* lexical variables *)
  "a=1;puts Array.new(10){|x| x*a }.reduce(0){ |x,y| x+y; }";

  (* function *)
  "def m(x) x*1; end; puts Array.new(10){|x| m(x)}.reduce(0){ |x,y| x+y }";

  (* function called inside of a function *)
  "def m(x) n(x*1); end; def n(x) x; end; puts Array.new(10){|x| m(x)}.reduce(0){ |x,y| x+y }";

  (* mutually recursive functions, which requires prototype generation *)
  "\
def m(x) if x==0 then 0 else n(x-1) end; end
def n(x) if x==0 then 1 else m(x-1) end; end
puts Array.new(10){|x| m(x)}.reduce(0){ |x,y| x+y }";

  (* functions both called and not called from kernel *)
  "\
def m(x) x; end
def n(x) puts x; end
n(Array.new(10){|x| m(x)}.reduce(0){ |x,y| x+y })";

  (* a function called from two locations *)
  "\
def m(x) x; end
puts Array.new(10){|x| m(x)}.reduce(0){ |x,y| m(x+y) }";

  (* multi-demensional array *)
  "puts(Array.new_(10,10){ |i,j| i*10 + j }.reduce(0){ |x,y| x+y })";

  (* multi-demensional array *)
  "puts(Array.new_(10,10){ |i,j| i*10 + j }.\
   map{ |x| x+1 }.reduce(0){ |x,y| x+y })";

  (* different input output types in reduce *)
  "puts Array.new(10){|x| x}.reduce(0){ |x,y| x+y+0.5 }";

  (* zip *)
  "puts(Array.new(10){|x| x+1}.zip(Array.new(10){|x|x}).map{ |x,y| x-y }.
        reduce(0){ |x,y| x+y })";

  (* zip with three arrays *)
  "a=Array.new(10){|x| x };
   b=a.zip(a,a).map{ |a,b,c| a+b+c };
   c=b.reduce(0){ |x,y| x+y };
   puts c";

  (* nested zip *)
  "a=Array.new(10){|x| x };
   b=Array.new(10){|x| x*x };
   c=Array.new(10){|x| x+x+1 };
   d=a.zip(b).zip(c).map{ |(a0,b0),c0| a0+b0/c0 };
   puts d.reduce(0){ |x,y| x+y }";

  (* blocks less number of parameters *)
  (* TODO: Ruby allows blocks to provide less number of formal
     parameters, which is not supported yet. *)
  (* "a=Array.new(10){|x| x};\ *)
  (*  a.zip(a).map{ |v| v*v }" *)

  (* zip with 2D arrays *)
  "a=Array.new_(5,5){ |x,y| x*5+y };\
   b=a.zip(a).map{ |x,y| x+y };\
   puts b.reduce(0){ |x,y| x+y }";

  (* neighbor *)
  "puts Array.new(10){ |x| x }.neighbor3(0).map{ |n| n[-1]*n[0]*n[1] }\
        .reduce(0){ |x,y| x+y }";

  (* neighbor9 *)
  "puts Array.new_(10,10){ |x,y| x*10+y }\
             .neighbor9(0).map{ |n| (n[-1,-1] + n[-1,0] + n[1,1] +\
                                     n[ 0,-1] + n[ 0,0] + n[0,1] +\
                                     n[ 1,-1] + n[ 1,0] + n[1,1]) / 9
             }.reduce(0){ |x,y| x+y }";

  (* neighbor27 *)
  "puts Array.new_(3,3,3){ |x,y,z| x*9+y*3+z }\
             .neighbor27(0).map{ |n|
                        (n[-1,-1,-1]*10000 + n[0,0,0]*100 + n[1,1,1])/3
             }.reduce(0){ |x,y| x+y }";

  (* lexical variable? *)
  " def m(x) Array.new(10){ |i| x*i }; end
    puts m(123).reduce(0){|x,y| x+y }";

  (* lexical variable *)
  " def m(x) a=x; Array.new(10){ |i| a*i }; end
    puts m(123).reduce(0){|x,y| x+y }";

  (* constant support *)
  "C=1; a = Array.new(C){ |x| x }";
  "C=1; a = Array.new(10){ |x| x*C }";
  "C=1; def m(x) x*C; end; a = Array.new(10){ |x| m(x) }";

  (* abs and floor *)
  "puts Array.new(10){ |x| (x*0.2).floor }.reduce(0){ |x,y| x+y }";
  "puts Array.new(10){ |x| (x-5).abs }.reduce(0){ |x,y| x+y }";
  "puts Array.new(10){ |x| (x-5.5).abs }.reduce(0){ |x,y| x+y }";

  (* function with global variables *)
  "SIZE=2
   def m() 1 end
   puts Array.new(1){ |x| m() }.reduce(0){ |x,y| x+y }";

  (* if-then-without-else in a method *)
  "def encodeHSBcolor() r1 = 0; if 0 < 1 then r1 = 0; end; r1 end
   a = Array.new_(10){|x| encodeHSBcolor()}";

  (* multiple statements in if-then-else *)
  "def encodeHSBcolor() if 0 < 1 then r1 = 0; r1 = 0; else 0; end; r1; end
   a = Array.new_(10){|x| encodeHSBcolor()}";

  (* floating modulo operator *)
  "def encodeHSBcolor() 3.23 % 2; end
   a = Array.new_(10){|x| encodeHSBcolor()}";

  (* power of negative numbers *)
  "puts Array.new_(1){ |x| (-10.0).abs**2}.reduce(0){|x,y| x+y }";

  (* support for boolean method in the kernel *)
  "def zero(x) x==0 end
   puts Array.new_(2){ |x| zero(x) ? 1 : 2  }.reduce(0){|x,y| x+y }";

  (* local variable inside a conditional branch  *)
  "def zero() if 0==0 then x=0; x=x+1 else 1 end  end
   puts Array.new_(2){ |x| zero()  }.reduce(0){|x,y| x+y }";

  (* true and false *)
  "def zero(x) if x==0 then false else true end  end
   puts Array.new_(2){ |x| zero(x) ? 1*x : 2*x  }.reduce(0){|x,y| x+y }
   puts zero(0) ? 1 : 2";

  (* for in Ruby *)
  "for i in 1..10 do; puts i; end";
  "c=10;for i in 1..c do; puts i; end";

  (* for in kernel *)
  "Array.new_(3){ |x| s=0; for i in 1..x do; s+=i; end; s }.\
 reduce(0){|x,y| x+y }";
  "Array.new_(3){ |x| s=0; range=1..x; for i in range do; s+=i; end; s }.\
 reduce(0){|x,y| x+y }";
  "Array.new_(3){ |x| s=0; range=1..x; for i in range do; s+=i; end; s }.\
 reduce(0){|x,y| x+y }";
  "Array.new_(3){ |x| 
     s=0; 
     for i in 1..x do; 
       for j in 1..i do; 
         s+=i*j; 
       end; 
     end;
     s 
 }.reduce(0){|x,y| x+y }";

  (* reduction without argument *)
  "Array.new_(3){ |x| x }.reduce{|x,y| x+y }";

  (* &:name as a block parameter *)
  "Array.new_(3){ |x| x }.reduce(0,&:+)";
  "Array.new_(3){ |x| x}.reduce(&:+)";

  (* method call on the LHS of an arithmetic *)
  "def n() 2; end; puts Array.new(1){ |i| i }.reduce(0){ |x,y| n()+0 }";

  (* indirect inequality resolution needed *)
  "def encodeHSBcolor(h) 1.0*(h - 1); end; \
   puts Array.new_(100){ |x_pixel| \
     c = x_pixel; encodeHSBcolor(c*1.0)\
   }.reduce(0){ |x,y| x+y }";

  (* return statement *)
  "def isprime(n)
     for i in 2..n-1
       if n%i == 0 then return 0; end
     end
     1
   end
   puts Array.new_(6){ |n| isprime(n) }.reduce(0){ |x,y| x+y }";

  (* return statement with more specific type *)
  "def m(n)
     if n%2==0 then return 0.5; end
     1
   end
   puts Array.new_(5){ |n| m(n) }.reduce(0){ |x,y| x+y }";

  (* array of tuples *)
  "puts Array.new(1){|i| [i+1,i+2]}.map{ |x,y| x*y }.reduce(0){ |a,b| a+b }";
  "puts Array.new(1){|i| i }.map{ |i| [i+1,i+2] }.map{ |x,y| x*y }.
              reduce(0){ |a,b| a+b }";

]

let test_tuple = [
  (* multiple assignments in kernel *)
  "Array.new_(3){ |x| a,b = [x,x/2]; a-b }.reduce(0){|x,y| x+y }";

  (* multiple assignments *)
  "a=1;b=2;a,b=[b,a];puts a";

  (* tuples in Kernel *)
  "Array.new(10){ |x| a=[1,x]; a[1] }.reduce(0){|x,y| x+y }";

  (* tuples in Ruby *)
  "a=[1,2];puts a[1]";
]

let test_procedure_as_symbol = [

  (* &:name as a block parameter *)
  "Array.new_(3){ |x| -x }.map(&:abs).reduce(&:+)";

  (* we need to know Object.to_f()->Float to solve the following
     case *)
  "Array.new_(3,&:to_f).map{ |x| x/2 }.reduce(&:+)#pending";


]

let test_scripts_float = [
  (* math support *)
  (* NOTE: results of these tests are compared with some precision *)
  "puts Array.new(10){ |x| Math::PI*x }.reduce(0){|x,y| x+y }";
  "puts Array.new(10){ |x| Math.cos(x) }.reduce(0){|x,y| x+y }";
]

let test_read_lines _ =
  let str = "Hello,\nworld." in
  Test_util.with_string
    (fun filename -> assert_equal str (Util.read_lines filename))
    str

let run_ruby_string : string -> string =
  fun script -> 
    Test_util.with_string
      (fun filename -> (Util.run_ruby ["-I"; "../proto/parray/simulator/";
				       "-r"; "parray";
				       filename]))
      script

let test_run_ruby _ =
  assert_equal "Hello." (run_ruby_string "puts 'Hello.'")

let compare_float_strings str1 str2 =
  let floats_of str =
    List.map float_of_string (Str.split (Str.regexp "\n") str) in
  List.for_all2 (cmp_float ~epsilon:1e-7) (floats_of str1) (floats_of str2)

let test_frontend ?(cmp=(=)) script =
  ("test_frontend("^script^")")>::
    fun _ -> assert_equal ~printer:(fun x->x) ~cmp
      (run_ruby_string script)
      (Frontend.compile_and_run script)

let suite =
  "suite">:::
    List.map test_frontend test_scripts
  @ if do_test_tuple then List.map test_frontend test_tuple else []
  @ if do_test_procedure_as_symbol
    then List.map test_frontend test_procedure_as_symbol else []
  @ (List.map (test_frontend ~cmp:compare_float_strings) test_scripts_float)
  @ [
      "test_run_ruby"  >:: test_run_ruby;
      "test_read_lines">:: test_read_lines;
    ];;

let _ = Config.initialize None

let () =
  run_test_tt_main suite
;;
