let _ = Config.initialize None

let gettime _ = Unix.gettimeofday()
let printtime _ = Printf.printf "time=%f\n" (gettime ())

let repeat_count = 10

let time name f x =
  let t = gettime () in
  let fx = f x in
  Printf.printf "time for %s: %fs\n" name ((gettime ()) -. t);
  fx

let source_string =
  "a = Array.new(10){|x| x }; p a.reduce(0){|x,y| x+y }"

let parse = Parser.parse_file

let string_of = Ast.string_of

let write_file  = Util.with_temp_file "parser" ".rb" 


let dummy_write_file source = 
  write_file source (fun x -> ())

let measure name f =
  time name
    (fun x ->
      let rec loop n =
	if n > 1 then
	  (f x; loop (n-1))
	else
          f x in
      loop repeat_count)


let _ = measure "with_temp_file" dummy_write_file source_string

let _ = write_file source_string (fun filename ->
  let sexp_str = (measure "ruby_parse_file" Parser.ruby_parse_file
		    filename) in
  let sexp = 	(measure "parse_sexp_string" Parser.parse_sexp_string
		   sexp_str) in
  let parsed = 	(measure "ast_of_sexp" Parser.ast_of_sexp sexp) in
  let str =      (measure "string_of" string_of parsed) in 
 (print_string str))



(* let _ =  *)

(*   (source *)
(*      (fun filename -> *)
(*        print_string (string_of (parse filename)))) *)
