open OUnit2
open Tarjan

let elements = [1;2;3;4;5;6;7;8;]
let relations = 
  [1, 2;
   2, 3;
   3, 1;
   4, 2; 4, 3; 4, 5;
   5, 4; 5, 6;
   6, 3; 6, 7;
   7, 6;
   8, 8; 8, 5;]
let expected = 
    [
      [8];
      [4;5];
      [6;7];
      [1;2;3];
    ]

let vertices = List.map make_node elements 

let _ = List.map (fun (n1, n2) ->
  connect (List.nth vertices (n1-1)) (List.nth vertices (n2-1)))
  relations

let expected_sccs =
  List.map (List.map (fun n -> (List.nth vertices (n-1))))
    expected

(* a test case for an internal function *)
let case1 _ =				
  let (sccs,index,stack) = tarjan vertices in
  assert_equal [] stack;
  assert_equal ~printer:string_of_sccs
    ~cmp:(fun sccs1 sccs2 ->
      List.for_all (fun (scc1,scc2)->
	List.for_all (fun ({element=e1},{element=e2}) -> e1=e2)
	  (List.combine scc1 scc2 ))
	(List.combine sccs1 sccs2))
    expected_sccs sccs 

(* a test case for an external function *)
let case2 _ = 
  assert_equal expected (Tarjan.sccs elements relations)

let suite = "suite_tarjan">:::
  [
    "wikipedia">::case1;
    "wrapper">::  case2;
  ]

let () =
  run_test_tt_main suite

  
