open OUnit2


let testdata_convert = [

  (* the body of a PArray call will be converted to a call on a kernel
     method. *)
  "PArray.new(10){ |x| _typedecl(Integer, :x); x }",
  "PArray.new(10){ |ptr,size| Ikra_Kernel.initialize0(ptr,size) }";

  (* the kernel method call in the converted code should have 
     lexical variables in the body as parameters. *)
  "_typedecl(Integer, :l); l=3; \
   PArray.new(10){ |x| _typedecl(Integer, :x); x+l }",
  "_typedecl(Integer, :l); l=3; \
   PArray.new(10){ |ptr,size| Ikra_Kernel.initialize0(ptr,size,l) }";

  (* the lexical variables sholud be sorted in the alphabetical order. *)
  "_typedecl(Integer, :l, :a, :b); b=4; l=3; a=9;\
   PArray.new(10){ |x| _typedecl(Integer, :x); a+x+l+b }",
  "_typedecl(Integer, :l, Integer, :a, Integer, :b); b=4; l=3; a=9;\
   PArray.new(10){ |ptr,size| Ikra_Kernel.initialize0(ptr,size,a,b,l) }";

  (* reduce calls on an array. *)
  "_typedecl([Array,1,Integer], :a); a=0; a.reduce(0){ |x,y| x+y }",
  "_typedecl([Array,1,Integer], :a); a=0; \
   a.reduce(0){ |ptr,size,zero| Ikra_Kernel.reduce0(ptr,size,zero)}";

  (* inject is synonym of reduce *)
  "_typedecl([Array,1,Integer], :a); a=0; a.inject(0){ |x,y| x+y }",
  "_typedecl([Array,1,Integer], :a); a=0; \
   a.reduce(0){ |ptr,size,zero| Ikra_Kernel.reduce0(ptr,size,zero)}";

  (* map on an array *)
  "_typedecl([Array,1,Integer], :a); a=0; a.map{ |x| x*x }",
  "_typedecl([Array,1,Integer], :a); a=0; \
   a.map{ |ptr,size| Ikra_Kernel.map0(ptr,size)}";

  (* combination of new and map. *)
  "_typedecl([Array,1,Integer], :a, Integer, :l, :m);\
   l = m = 1;\
   a=PArray.new(10){ |x| _typedecl(Integer, :x); x+l };\
   puts a.reduce(1){ |x,y|_typedecl(Integer, :x, :y); x*y*m }",
  "_typedecl([Array,1,Integer], :a, Integer, :l, Integer, :m);\
   l = m = 1;\
   a=PArray.new(10){ |ptr,size| Ikra_Kernel.initialize0(ptr,size,l) };\
   puts a.reduce(1){ |ptr,size,zero| Ikra_Kernel.reduce1(ptr,size,zero,m) }";
]

(* let test_convert_exps = List.map make_test_convert testdata_convert *)

let test_tast_recognize _ =
  assert_equal 
    (Tast.recognize (Test_util.parse_ruby_string "a=0"))
    (let open Tast in
	(Local_assign("a",(Literal(Ast.Integer_literal 0), Ikra_type.Tobject)),
	 Ikra_type.Tobject))

let test_tast_recognize2 _ =
  assert_equal 
    (Tast.recognize (Test_util.parse_ruby_string "_t(a=_t(0,Integer),Integer)"))
    (let open Tast in
	(Local_assign("a",(Literal(Ast.Integer_literal 0), Ikra_type.Tint)),
	 Ikra_type.Tint))

let test_tast_recognize3 _ =
  let open Tast in
  match 
    Tast.recognize (Test_util.parse_ruby_string "\
   _typedecl([Array, 1, Integer],:a);\
   a=PArray.new(10){ |x| _typedecl(Integer, :x); x };\
   puts _t(a.map{ |x| _typedecl(Integer, :x); x+1 },[Array,1,Integer]).\
          reduce(0){ |x,y| _typedecl(Integer, :x, :y); x+y }") with
      | Block(_,[_;_;Call(_,"puts",[Call(Some(_,t),"reduce",_,_),_],_),_]),_ ->
	assert_equal ~printer:Ikra_type.string_of 
	  (Ikra_type.Tarray (1, Ikra_type.Tint)) t
      | e -> invalid_arg ("unexpected."^ (Tast.string_of e))

let test_tast_recognize4 _ =
  (* check whether the return type declaration for a block can be
     recognized. *)
  let open Tast in
      match 
	Tast.recognize
	  (Typeinf.recognize_type_decls
	     (Test_util.parse_ruby_string "\
   PArray.new(10){ |x| _typedecl(Integer, :x, Float, :_RETURN); x }")) with
	    | Call(_,"new",_,Some(Block _,typ)),_ ->
	      assert_equal ~printer:Ikra_type.string_of 
		Ikra_type.Tfloat typ
	    | e -> invalid_arg ("unexpected."^ (Tast.string_of e))

let test_tast_recognize5 _ =
  let open Tast in
      match 
	Tast.recognize
	  (Typeinf.recognize_type_decls
	     (Test_util.parse_ruby_string "\
PArray.new(10){ |x| _typedecl(Integer, :x, Float, :_RETURN); x }")) with
	    | Call(_,"new",_,Some(Block(Var_pattern(x,typ), _),_)),_ ->
	      assert_equal ~printer:Ikra_type.string_of 
		Ikra_type.Tint typ
	    | e -> invalid_arg ("unexpected."^ (Tast.string_of e))


let suite =
  "suite_convert">:::[
    "TAST_recognize">::test_tast_recognize;
    "TAST_recognize2">::test_tast_recognize2;
    "TAST_recognize3">::test_tast_recognize3;
    "TAST_recognize4">::test_tast_recognize4;
    "TAST_recognize5">::test_tast_recognize5;
  ]

let () =
  run_test_tt_main suite

